#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <syslog.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/resource.h>	      
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include "dhcp.h"
#include "util.h"

static struct dhcp_config cfg;

void do_log(int priority, char * format, ...)
{
	va_list ap;
	char maxlog[1024];
	
	va_start(ap, format);
	vsnprintf(maxlog, 1024, format, ap);
	
	if(cfg.syslog)
		syslog(priority, "%s", maxlog);
	else
	{
		time_t t;
		char * littime;
		char * r;
		t = time(NULL);
		littime = ctime(&t);
		if((r = index(littime, '\n'))) *r = 0;
		fprintf(stderr, "%s %d %s\n", littime, priority, maxlog);
	}
	
	va_end(ap);
}

void wait_handler(int sig)
{
	while(wait3(NULL, WNOHANG, (struct rusage *) NULL) > 0);
}

void chld_handler(int sig)
{
}

void alarm_handler(int sig)
{
	do_log(LOG_NOTICE, "Request timed out, exiting.");
	exit(0xad);
}

inline double time_diff(struct timeval *s, struct timeval *e)
{
	double v;
	v = e->tv_usec  - s->tv_usec;
	v /= 1000000;
	v += e->tv_sec - s->tv_sec;
	return v;
}

int main(int argc, char ** argv)
{
	int listening;
	int sending = 0;
	int uid = 0, gid = 0;

	struct sockaddr_in peer_addr;
	
	char request[PACKETSIZE];
	int request_len;

	int t = 1;
	
	struct timeval t_start;
	struct timeval t_end;

	struct timeval t_lastreport;

	double t_request;
	double t_sum = 0;
	double t_select = 0;
	int N_requests = 0;
	int N_timeout = 0;
	
	fd_set fds;

	int c;

	memset(&cfg, 0, sizeof(cfg));
	cfg.syslog = 1;
	cfg.detach = 1;
	cfg.sin.sin_family = AF_INET;
	cfg.sin.sin_port = htons(67);
	cfg.sin.sin_addr.s_addr = INADDR_ANY;
	cfg.timeout = 10;
	cfg.nth = 0;
	cfg.account = 0;
	cfg.addr.sin_family = AF_INET;
	inet_aton("127.0.0.1", &cfg.addr.sin_addr);

	while((c = getopt(argc, argv, "h?t:efp:t:b:ai:U:G:x:P:A:")) != EOF)
	{
		switch(c)
		{
			case 'h':
			case '?':
				printf("autodhcp DHCP super-server (C) 2003 - 2006, 2013 Mateusz 'mteg' Golicz\n\n"
				       "Usage:\n"
				       "  autodhcp [-p <port>] [-t <timeout>] [-b <address>] [-e] [-f] (-x <command> | -P <port>)\n\n"
				       "Available options:\n"
				       " -p <port>    Port to bind to (default: 67 - standard DHCP)\n"
				       " -t <timeout> Request timeout in seconds (default: 10) - after this time, the child process is killed. It is really unwise to unset the timeout (0), as the DHCP server may process only one request at time.\n"
				       " -b <address> Bind the listening socket to this address (instead of 0.0.0.0)\n"
				       " -f           Do not go into the background\n"
				       " -a           Account and report request handling times\n"
				       " -i <count>   Print timing summary every <count> requests (implies -a)\n"
				       " -e           Output the log to the stderr instead of syslog\n"
				       " -A <ip>      IP address of helper service (default: 127.0.0.1)\n"
				       " -U <uid>     Drop privileges to <UID> after initialization\n"
				       " -G <gid>     Drop privileges to <GID> after initialization\n\n"
				       "If the -c option is specified, whenever the server receives a DHCP packet, <command>"
				       " is executed as a child process by calling /bin/sh -c <command>\n\n"
				       "Alternatively, the -P option can be specified to make autodhcp use a helper service running at the specified port\n" );
				return(1);
			case 'U':
				uid = parselong(optarg, "Invalid UID: %s\n", 0, 65535);
				break;
			case 'G':
				gid = parselong(optarg, "Invalid GID: %s\n", 0, 65535);
				break;
			case 'e':
				cfg.syslog = 0;
				break;
			case 'f':
				cfg.detach = 0;
				break;
			case 'p':
				cfg.sin.sin_port = htons(parselong(optarg, "Invalid port number: %s\n", 0, 65535));
				break;
			case 't':
				cfg.timeout = parselong(optarg, "Invalid timeout: %s\n", 0, 0);
				break;
			case 'a':
				cfg.account = 1;
				break;
			case 'i':
				cfg.account = 1;
				cfg.nth = parselong(optarg, "Invalid summary interval: %s\n", 1, 1000000);
				break;
			case 'P':
				cfg.addr.sin_port  = htons(parselong(optarg, "Invalid port number: %s\n", 1, 65536));
				break;
			case 'A':
				if(!inet_aton(optarg, &cfg.addr.sin_addr))
				{
					fprintf(stderr, "Incorrect IP address: %s\n", optarg);
					return 1;
				}
				break;
			case 'x':
				assert(cfg.command = strdup(optarg));
				break;
			
		}
	}

	if(argv[optind])
	{
		fprintf(stderr, "Bogus commandline arguments, starting with: %s\n", argv[optind]);
		return(1);
	}
	
	if(!(cfg.command || cfg.addr.sin_port))
	{
		fprintf(stderr, "Neither a command nor a port number was given. Please use either -P or -x.\n");
		return 1;
	}

	if((listening = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		fprintf(stderr, "Unable to create a listening socket: socket(): %s\n", strerror(errno));
		return(1);
	}

	if(bind(listening, (struct sockaddr*) &cfg.sin, sizeof(struct sockaddr_in)) < 0)
	{
		fprintf(stderr, "Unable to bind to %s:%d: bind(): %s. Ensure that I have the proper privileges.\n", inet_ntoa(cfg.sin.sin_addr), ntohs(cfg.sin.sin_port), strerror(errno));
		return(1);
	}
	cfg.dgramsock = listening;

	if(setsockopt(listening, SOL_SOCKET, SO_BROADCAST, &t, sizeof(int)) < 0)
	{
		fprintf(stderr, "Unable to setup broadcast receiving: setsockopt(): %s. Ensure that I have the proper privileges.\n", strerror(errno));
		return(1);
	}

	t = 1;
	if(setsockopt(listening, SOL_IP, IP_PKTINFO, &t, sizeof(int)) < 0)
	{
		fprintf(stderr, "Unable to setup interface information receiving: setsockopt(): %s. Ensure that I have the proper privileges.\n", strerror(errno));
		return(1);
	}

	if(gid)
		if(!setregid(gid, gid))
		{
			fprintf(stderr, "Cannot drop privileges to GID %d: %s\n", gid, strerror(errno));
			return 1;
		}

	if(uid)
		if(!setreuid(uid, uid))
		{
			fprintf(stderr, "Cannot drop privileges to UID %d: %s\n", uid, strerror(errno));
			return 1;
		}
	
	if(cfg.detach)
	{
		switch(fork())
		{
			case 0:
				setsid();
				break;
			case -1:
				perror("fork");
				break;
			default:
				exit(0);
		}
	}
	
	
	signal(SIGCHLD, chld_handler);
	do_log(LOG_DEBUG, "We are good to go!");
	gettimeofday(&t_lastreport, NULL);
	while(1)
	{
		struct dhcp_requesthdr * r = (struct dhcp_requesthdr*) request;
		char control[PACKETSIZE];
		struct msghdr msgh;
		struct iovec iv;
		struct cmsghdr *cmsg;
		
		struct in_addr local;
		int ifi = -1;
		int n_fd;
		
		struct timeval select_tv;

		
		do
		{
			FD_ZERO(&fds);
			FD_SET(listening, &fds);

			select_tv.tv_sec = 10;
			select_tv.tv_usec = 0;
			n_fd = select(listening + 1, &fds, NULL, NULL, &select_tv);
			if(n_fd >= 0)
			{
				t_select += 10 - select_tv.tv_sec;
				t_select += (0 - select_tv.tv_usec) / 1000000.0;
			}
		} while(n_fd == 0);
		
		iv.iov_base = request;
		iv.iov_len = PACKETSIZE;
		
		msgh.msg_name = &peer_addr;
		msgh.msg_namelen = sizeof(struct sockaddr_in);
		msgh.msg_iov = &iv;
		msgh.msg_iovlen = 1;
		msgh.msg_control = control;
		msgh.msg_controllen = PACKETSIZE;

		if((request_len = recvmsg(listening, &msgh, 0)) <= 0)
		{
			do_log(LOG_ERR, "recvfrom(): %s. Terminating.", strerror(errno));
			return 1;
		}
		
		local.s_addr = 0;
		
		for(cmsg = CMSG_FIRSTHDR(&msgh); cmsg; cmsg = CMSG_NXTHDR(&msgh, cmsg))
		{
			if(cmsg->cmsg_level == SOL_IP && cmsg->cmsg_type == IP_PKTINFO)
			{
				struct in_pktinfo * pi;
				pi = (struct in_pktinfo*) CMSG_DATA(cmsg);
				if((cmsg->cmsg_len - sizeof(struct cmsghdr)) >= sizeof(struct in_pktinfo))
				{
					ifi = pi->ipi_ifindex;
					local = pi->ipi_spec_dst;
					break;
				}
			}
		}
                      
//		do_log(LOG_NOTICE, "Interface: %d local address %s.", ifi, inet_ntoa(local));


		if(msgh.msg_namelen != sizeof(struct sockaddr_in))
		{
			do_log(LOG_NOTICE, "Invalid peer socket address received, unable to determine peer parameters. Perhaps I am incompatible with this system.");
			continue;
		}


		if(request_len < MIN_DHCP_REQUEST)
		{
			do_log(LOG_NOTICE, "Received a message that is shorter (%d bytes) than the minimal DHCP request size, ignoring.", request_len);
			continue;
		}
		if(r->op != OP_BOOTREQUEST)
		{
			do_log(LOG_NOTICE, "Received a message that is not a DHCP request (opcode: %d), ignoring.", r->op);
			continue;
		}
		if(r->hlen != ETH_ALEN)
		{
			do_log(LOG_NOTICE, "Received a message that is not an ethernet DHCP request, ignoring.");
			continue;
		}

/*		do_log(LOG_NOTICE, "wait3_debug: Parent: my PID is %d, fork()", getpid());*/
		if(cfg.account)
			gettimeofday(&t_start, NULL);
		
		switch(fork())
		{
			case -1:
				do_log(LOG_ERR, "Unable to handle a request: fork(): %s", strerror(errno));
				break;
			case 0:
/*				do_log(LOG_NOTICE, "wait3_debug: Child: Starting up! My PID is %d...", getpid());*/
				if(cfg.timeout)
					alarm(cfg.timeout);
				signal(SIGALRM, alarm_handler);
				signal(SIGCHLD, wait_handler);
				dhcp_handle(r, peer_addr, request_len, sending, &cfg, ifi, local);
/*				do_log(LOG_NOTICE, "wait3_debug: Child %d: bye!...", getpid());*/
				exit(0);
			default:
			{
				int k;
				do {
					int status;
					k = wait3(&status, 0, (struct rusage *) NULL);
					
					if(WIFEXITED(status))
					{
						if(WEXITSTATUS(status) == 0xad)
							N_timeout ++;
					}
					
/*					do_log(LOG_NOTICE, "wait3_debug: Parent: wait3() = %d, %s", k, strerror(errno));*/
				} while(k > 0);
				break;
			}
		}

		if(cfg.account)
		{
			gettimeofday(&t_end, NULL);
			t_request = time_diff(&t_start, &t_end);
			do_log(LOG_NOTICE, "Request handling time: %.2f s\n", t_request);
			if(cfg.nth)
			{
				t_sum += t_request;
				N_requests++;
				if(N_requests >= cfg.nth)
				{
					double t_sincelast;
					t_sincelast = time_diff(&t_lastreport, &t_end);
					memcpy(&t_lastreport, &t_end, sizeof(struct timeval));
					do_log(LOG_NOTICE, "::: Summary: Last %d requests: overall elapsed %.1f s; total idle time (waiting in select) %.1f s; average handling time %.2f s; %d requests had timed out\n", N_requests, t_sincelast, t_select, t_sum / N_requests, N_timeout);
					t_sum = 0;
					t_select = 0;
					N_requests = 0;
					N_timeout = 0;
				}
			}
		}

	}
	
	do_log(LOG_ERR, "select(): %s. Terminating.", strerror(errno));
	return 1;
}
