#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <syslog.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include "dhcp.h"
#include "dhcpparse.h"


int send_message(char *mac, int sending, struct dhcp_config *cfg, u_int8_t *message, int len, struct in_addr srcip, int ifi, struct in_addr local)
{
	int res;
	u_int32_t relay, dst;
	
	relay = *((u_int32_t*) (((struct dhcp_requesthdr*) message)->giaddr));
	dst = *((u_int32_t*) (((struct dhcp_requesthdr*) message)->yiaddr));
	
	if(relay)
	{
		struct sockaddr_in sin;
		
		sin.sin_family = AF_INET;
		sin.sin_port = htons(67);
		sin.sin_addr.s_addr = relay;
		res = sendto(cfg->dgramsock, message, len, 0, (struct sockaddr*) &sin, sizeof(struct sockaddr_in));
		return(res);
	}

	
	if(ifi < 0)
	{
		do_log(LOG_NOTICE, "Message from %s: Unable to determine destination interface. Will not respond to the message despite having the response present and accounted for.", mac);
		return(0);
	}
	
	if(!srcip.s_addr)
		srcip = local;

	struct sockaddr_in to;
	
	to.sin_family = AF_INET;
	to.sin_port = htons(68);
	to.sin_addr.s_addr = dst;
	res = sendto(cfg->dgramsock, message, len, 0, (struct sockaddr*) &to, sizeof(struct sockaddr_in));
	return(res);
}

int exchange(int fd, struct dhcp_config *c, char *in, int len, char *out, int maxlen, char *mac)
{
	int inpos = 0, outpos = 0;
	
	while(1)
	{
		fd_set rfds;
		fd_set wfds;
		
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);

		FD_ZERO(&wfds);
		
		if(inpos < len)
			FD_SET(fd, &wfds);
		
		if(select(fd + 1, &rfds, &wfds, NULL, NULL))
		{
			if(FD_ISSET(fd, &wfds))
			{
				int l;
				l = write(fd, in + inpos, len - inpos);
				if(l < 0)
				{
					do_log(LOG_NOTICE, "Message from %s: write(): %s.", mac, strerror(errno));
					close(fd);
					return(1);
				}
				inpos += l;
				if(inpos >= len)
					shutdown(fd, SHUT_WR);
			}
			if(FD_ISSET(fd, &rfds))
			{
				int l;
				l = read(fd, out + outpos, maxlen - outpos);
				if(l < 0)
				{
					do_log(LOG_NOTICE, "Message from %s: read(): %s.", mac, strerror(errno));
					close(fd);
					return(1);
				}
				if(l == 0)
				{
					out[outpos] = 0;
					shutdown(fd, SHUT_RDWR);
					close(fd);
					return(0);
				}
				outpos += l;
			}
		}
	}
	close(fd);
}

int xchg_info(struct dhcp_config * c, char * in_buffer, int len, char * out_buffer, int maxlen, char *mac)
{
	if(c->command)
	{
		int pipes[2];
		in_buffer[len] = 0;

		if(socketpair(AF_UNIX, SOCK_STREAM, 0, pipes) < 0)
		{
			do_log(LOG_NOTICE, "Message from %s: socketpair(): %s", mac, strerror(errno));
			return(1);
		}

		switch(fork())
		{
			case -1:
				do_log(LOG_NOTICE, "Message from %s: fork(): %s", mac, strerror(errno));
				return(1);
			case 0:
				if(c->timeout)
					alarm(c->timeout);
				dup2(pipes[1], 1);
				dup2(pipes[1], 0);
				execl("/bin/sh", "/bin/sh", "-c", c->command, NULL);
				do_log(LOG_NOTICE, "Message from %s: execl(): %s", mac, strerror(errno));
				_exit(1);			
			default:
				close(pipes[1]);
				return(exchange(pipes[0], c, in_buffer, len, out_buffer, maxlen, mac));
		}
	}
	else
	{
		int s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if(s == -1)
		{
			do_log(LOG_NOTICE, "Message from %s: socket(): %s", mac, strerror(errno));
			return(1);
		}
		if(connect(s, (struct sockaddr_in*) &c->addr, sizeof(struct sockaddr_in)))
		{
			do_log(LOG_NOTICE, "Message from %s: connect(): %s", mac, strerror(errno));
			return(1);
		}
		return(exchange(s, c, in_buffer, len, out_buffer, maxlen, mac));
	}
	return(1);
}


void dhcp_handle(struct dhcp_requesthdr *r, struct sockaddr_in peer, int length, int sending, struct dhcp_config * cfg, int ifi, struct in_addr local)
{
	char mac[20];
	
	u_int8_t *options;
	int optlen, offs;
	
	snprintf(mac, 20, "%02x:%02x:%02x:%02x:%02x:%02x", r->chaddr[0],
		r->chaddr[1], r->chaddr[2], r->chaddr[3], r->chaddr[4], r->chaddr[5]);


//	do_log(LOG_INFO, "Message from %s: Received.", mac);

	if(ntohl(*((u_int32_t*) (r->magic))) != DHCP_MAGIC)
	{
		do_log(LOG_INFO, "Message from %s: Ignoring, invalid magic number (%04x)", mac, *((u_int32_t*) (r->magic)));
		return;
	}
	
	options = ((char*) r) + sizeof(struct dhcp_requesthdr);
	optlen = length - sizeof(struct dhcp_requesthdr);

	if(optlen < 1)
	{
		do_log(LOG_INFO, "Message from %s: Ignoring, message unconsistent (no options).", mac);
		return;
	}

	for(offs = 0; offs < optlen; )
	{
		if(!options[offs])
			offs++;
		else if(options[offs] == 0xff)
			break;
		else
			offs += options[offs + 1] + 2;
	}
	
	if(offs > optlen)
	{
		do_log(LOG_INFO, "Message from %s: Ignoring, message unconsistent (inconsistent option set).", mac);
		return;
	}
	
	offs = dhcp_find_option(options, optlen, DHCP_OPTION_TYPE);
	if(offs < 0)
	{
		do_log(LOG_INFO, "Message from %s: Ignoring, message unconsistent (no message type option present).", mac);
		return;
	}

	{
		char message[PACKETSIZE];
		int len, res;
		struct in_addr ina;
		
		char in_buffer[DESCRIPTION_BUFFER + 12];
		char out_buffer[DESCRIPTION_BUFFER + 12];
	
//		do_log(LOG_INFO, "Message from %s: Describing.", mac);
		
				
		len = dhcp_describe(r, peer, length, in_buffer, DESCRIPTION_BUFFER, mac);
//		do_log(LOG_INFO, "Message from %s: Description complete.", mac);
		
		if(len < 0)
		{
			do_log(LOG_INFO, "Message from %s: Unable to parse the message, aborting.", mac);
			return;
		}
		
		do_log(LOG_INFO, "Message from %s: Retrieving response.", mac);
		if(xchg_info(cfg, in_buffer, len, out_buffer, DESCRIPTION_BUFFER, mac))
		{
			do_log(LOG_INFO, "Message from %s: Unable to retrieve a response, aborting", mac);
			return;
		}
		len = dhcp_build(out_buffer, (struct dhcp_requesthdr*) message, r, PACKETSIZE, &ina);

		if(len < 0)
		{
			do_log(LOG_INFO, "Message from %s: Unable to convert the response into a reply message, aborting", mac);
			return;
		}

		do_log(LOG_INFO, "Message from %s: Sending response (%d bytes).", mac, len);
		
		res = send_message(mac, sending, cfg, message, len, ina, ifi, local);
//		res = send_message(sending, r, *((u_int32_t*) msg->yiaddr), cfg, message, len, net->ifi, net->srcip ? net->srcip : cfg->srcip);
		if(res < 0)
			do_log(LOG_NOTICE, "Message from %s: sendto(): %s.", mac, strerror(errno));

		return;
	}
	
//	do_log(LOG_INFO, "Message from %s: Ignoring, message unconsistent (bad options).", mac);
	return;
}
